/**
 * @file rudp-light.c
 *
 * @brief Reliable User Datagram Protocol (RUDP) - minimal implemenation.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: rudp-light.c 2565 2008-02-20 12:57:20Z wittenbu $
 */

#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "../types.h"
#include "../vars.h"
#include "../interface.h"
#include "../interface_internal.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "rudp-light.h"
#include "rudp-light_types.h"
#include "rudp-light_vars.h"

#define SP_TYPE_RUDP_DATA	0x1	///< RUDP DATA packet type.
#define SP_TYPE_RUDP_ACK	0x2	///< RUDP ACK packet type.

/**
 * @brief Maximal transmission queue length.
 */
#define RUDP_MAX_QUEUED_PACKETS	128

/**
 * @brief Number of transmission retries.
 */
#define RUDP_PACKET_RETRIES	5

/**
 * @brief Base of the exponential backoff time function.
 */
#define RUDP_BACKOFF_TIME_BASE	2

/**
 * @brief Minimal backoff time.
 */
#define RUDP_MIN_BACKOFF_TIME	SP_EXPECTED_MULTI_HOP_RTT

/**
 * @brief Maximal backoff time.
 */
#define RUDP_MAX_BACKOFF_TIME	SP_ROUTE_ESTABLISHMENT_TIME


/*
 * Interface to main module.
 */

/**
 * @brief Initialize the RUDP module.
 */
void RUDP_init() {
	list_alloc(&SP_RUDP_transmission_queue);
	SP_RUDP_total_transmissions = 0;
	SP_RUDP_successful_transmissions = 0;
	SP_RUDP_failed_transmissions = 0;
}

/**
 * @brief Start the RUDP module.
 */
void RUDP_start() {
}

/**
 * @brief Stop the RUDP module.
 */
void RUDP_stop() {
}

/**
 * @brief Finalize the RUDP module.
 */
void RUDP_finalize() {
	if(!list_empty(SP_RUDP_transmission_queue)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "RUDP: There are still %d packet(s) in the transmission queue.\n", list_size(SP_RUDP_transmission_queue));
		abort();
	} else if(SP_RUDP_total_transmissions != (SP_RUDP_successful_transmissions + SP_RUDP_failed_transmissions)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "RUDP: Transmit counters don't add up (%d != (%d + %d)).\n", SP_RUDP_total_transmissions, SP_RUDP_successful_transmissions, SP_RUDP_failed_transmissions);
		abort();
	}
	list_free(SP_RUDP_transmission_queue);
}



/*
 * Packet handling functions.
 */

/**
 * @brief Send a RUDP DATA packet.
 *
 * @param orig_header Pointer to header of the packet to be transmitted.
 * @param orig_payload Pointer to payload of the packet to be transmitted.
 */
static void send_RUDP_DATA(const SP_packet_header_t* orig_header, const SP_packet_payload_t* orig_payload) {
	SP_packet_header_t rudp_header;
	SP_RUDP_DATA_packet_t rudp_data;

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "RUDP: Sending DATA packet to node %s (seqno=%d)...\n", addr_to_str(orig_header->to), orig_header->seqno);
	memcpy(&rudp_header, orig_header, sizeof(SP_packet_header_t));
	rudp_header.app = SP_APP_RUDP;
	rudp_header.type = SP_TYPE_RUDP_DATA;
	rudp_header.payload_length = orig_header->payload_length + sizeof(SP_RUDP_DATA_packet_t);
	rudp_data.orig_app = orig_header->app;
	rudp_data.orig_type = orig_header->type;

	SP_packet_payload_t* rudp_payload;
	SP_MALLOC_JOIN(rudp_payload, SP_packet_payload_t, &rudp_data, sizeof(SP_RUDP_DATA_packet_t), orig_payload, orig_header->payload_length);
	SP_send(&rudp_header, rudp_payload);
	SP_FREE(rudp_payload);
}

/**
 * @brief Send a RUDP ACK packet.
 *
 * @param to Address of the destination node.
 * @param seqno Sequence number of the packet to be acknowledged.
 */
static void send_RUDP_ACK(SP_addr_t to, SP_seqno_t seqno) {
	SP_packet_header_t header;
	SP_RUDP_ACK_packet_t ack;

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "RUDP: Sending ACK for DATA packet with sequence number %d to node %s...\n", seqno, addr_to_str(to));
	init_header(&header);
	header.to = to;
	header.app = SP_APP_RUDP;
	header.type = SP_TYPE_RUDP_ACK;
	header.payload_length = sizeof(SP_RUDP_ACK_packet_t);
	ack.seqno = seqno;
	SP_send(&header, (SP_packet_payload_t*) &ack);
}

/**
 * @brief Handle incoming RUDP DATA packet.
 *
 * @param rudp_header Pointer to header of incoming RUDP DATA packet.
 * @param rudp_payload Pointer to payload of incoming RUDP DATA packet.
 */
static void handle_RUDP_DATA(const SP_packet_header_t* rudp_header, const SP_packet_payload_t* rudp_payload) {
	SP_RUDP_DATA_packet_t rudp_data;
	SP_packet_header_t orig_header;
	SP_packet_payload_t* orig_payload;

	if(rudp_header->payload_length < sizeof(SP_RUDP_DATA_packet_t)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "RUDP: Received DATA packet with wrong payload length (%d != %d).\n", rudp_header->payload_length, sizeof(SP_RUDP_ACK_packet_t));
		abort();
	}

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "RUDP: Received DATA from node %s (seqno=%d).\n", addr_to_str(rudp_header->from), rudp_header->seqno);
	send_RUDP_ACK(rudp_header->from, rudp_header->seqno);

	size_t orig_payload_length = rudp_header->payload_length - sizeof(SP_RUDP_DATA_packet_t);
	SP_MALLOC_ARRAY(orig_payload, SP_packet_payload_t, orig_payload_length);
	SP_SPLIT(&rudp_data, sizeof(SP_RUDP_DATA_packet_t), orig_payload, orig_payload_length, rudp_payload);
	memcpy(&orig_header, rudp_header, sizeof(SP_packet_header_t));
	orig_header.app = rudp_data.orig_app;
	orig_header.type = rudp_data.orig_type;
	orig_header.payload_length = orig_payload_length;
	if(orig_header.payload_length == 0)
		SP_handle_packet(&orig_header, NULL, true);
	else
		SP_handle_packet(&orig_header, orig_payload, true);
	SP_FREE(orig_payload);
}

/**
 * @brief Check whether destination address and sequence number of two
 * transmission match.
 *
 * @param item Transmission to check.
 * @param cmp_item Reference transmission.
 * @return <code>true</code> if destination address and sequence number of two
 * transmissions are equal, <code>false</code> otherwise.
 */
static bool matches_transmission_packet_header_to_and_seqno(const void* item, const void* cmp_item) {
	SP_RUDP_transmission_t* transmission = (SP_RUDP_transmission_t*) item;
	SP_RUDP_transmission_t* cmp_transmission = (SP_RUDP_transmission_t*) cmp_item;
	return (
		(transmission->packet.header->to == cmp_transmission->packet.header->to)
		&& (transmission->packet.header->seqno == cmp_transmission->packet.header->seqno)
	);
}

/**
 * @brief Handle incoming RUDP ACK packet.
 *
 * @param header Pointer to header of incoming RUDP ACK packet.
 * @param ack Pointer to ACK information of incoming RUDP ACK packet.
 */
static void handle_RUDP_ACK(const SP_packet_header_t* header, const SP_RUDP_ACK_packet_t* ack) {
	SP_RUDP_transmission_t cmp_transmission;
	SP_packet_header_t cmp_header;
	cmp_transmission.packet.header = &cmp_header;
	cmp_transmission.packet.header->to = header->from;
	cmp_transmission.packet.header->seqno = ack->seqno;
	SP_RUDP_transmission_t* transmission = (SP_RUDP_transmission_t*) list_find(SP_RUDP_transmission_queue, &cmp_transmission, matches_transmission_packet_header_to_and_seqno);
	if(transmission == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "RUDP: Ignoring ACK from node %s for packet with sequence number %d that is not in queue...\n", addr_to_str(header->from), ack->seqno);
		return;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "RUDP: Received ACK from node %s for packet with sequence number %d.\n", addr_to_str(header->from), ack->seqno);
	transmission->acked = true;
	if(transmission->function != NULL) {
		transmission->result->errno = SP_ERROR_OK;
		transmission->function(transmission->result);
	}
	transmission->result = NULL;
	list_remove(SP_RUDP_transmission_queue, transmission);
	if(transmission->packet.payload != NULL)
		SP_FREE(transmission->packet.payload);
	// Packet and header are freed in the timeout handler.
}

/**
 * @brief RUDP packet classifier.
 *
 * @param header Pointer to header of an incoming packet.
 * @param payload Pointer to payload of an incoming packet.
 */
void RUDP_recv(const SP_packet_header_t* header, const SP_packet_payload_t* payload) {
	if(header->app != SP_APP_RUDP) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "RUDP: Received non-RUDP packet.\n");
		abort();
	}

	switch(header->type) {
	case SP_TYPE_NONE:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "RUDP: Received packet with uninitialized type field from node %s.\n", addr_to_str(header->from));
		break;
	case SP_TYPE_RUDP_DATA:
		handle_RUDP_DATA(header, payload);
		break;
	case SP_TYPE_RUDP_ACK:
		if(header->payload_length != sizeof(SP_RUDP_ACK_packet_t)) {
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "RUDP: Received ACK packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SP_RUDP_ACK_packet_t));
			abort();
		}
		handle_RUDP_ACK(header, (SP_RUDP_ACK_packet_t*) payload);
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "RUDP: Received packet of unknown type 0x%X from node %s.\n", header->type, addr_to_str(header->from));
		abort();
	}
}



/*
 * Interface provided by this module.
 */

/**
 * @brief Timeout callback for sending a packet via RUDP.
 *
 * @param parameter Transmission for which the timeout ocurred.
 */
static void RUDP_send_timeout(SP_callback_parameter_t parameter) {
	SP_RUDP_transmission_t* transmission = (SP_RUDP_transmission_t*) parameter;

	if(transmission->acked) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "RUDP: Timeout for acknowledgement of DATA packet to node %d with sequence number %d has expired. Packet has been acknowledged.\n", addr_to_str(transmission->packet.header->to), transmission->packet.header->seqno);
		SP_FREE(transmission->packet.header);
		SP_FREE(transmission);

		SP_RUDP_successful_transmissions++;

	} else {
		// Retry if the packet hasn't been removed from the queue.
		if(transmission->retries > 0) {
			int retry = RUDP_PACKET_RETRIES - (--transmission->retries);
			SP_time_t backoff_time = SP_EXPECTED_MULTI_HOP_RTT * pow(RUDP_BACKOFF_TIME_BASE, retry);
			if(backoff_time > RUDP_MAX_BACKOFF_TIME)
				backoff_time = RUDP_MAX_BACKOFF_TIME;
			SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "RUDP: Timeout for acknowledgement of DATA packet to node %s with sequence number %d has expired (retry #%d). Resending packet and rescheduling by %.2fs...\n", addr_to_str(transmission->packet.header->to), transmission->packet.header->seqno, retry, backoff_time);
			send_RUDP_DATA(transmission->packet.header, transmission->packet.payload);
			callback_add(RUDP_send_timeout, transmission, backoff_time);

		// Remove from queue and drop packet if no retries are left.
		} else {
			SP_DEBUG(SP_DEBUG_LEVEL_WARN, "RUDP: Timeout for acknowledgement of DATA packet to node %s with sequence number %d has expired. No retries left, dropping packet...\n", addr_to_str(transmission->packet.header->to), transmission->packet.header->seqno);
			if(transmission->function != NULL) {
				transmission->result->errno = SP_ERROR_TIMEOUT;
				transmission->function(transmission->result);
			}
			list_remove(SP_RUDP_transmission_queue, transmission);
			SP_FREE(transmission->packet.header);
			if(transmission->packet.payload != NULL)
				SP_FREE(transmission->packet.payload);
			SP_FREE(transmission);

			SP_RUDP_failed_transmissions++;
		}
	}
}

/**
 * @brief Send a packet via RUDP.
 *
 * @param header Pointer to the header of the packet to send.
 * @param payload Pointer to the payload of the packet to send.
 * @param function Callback function.
 * @param result Result of the transmission.
 */
void RUDP_send(const SP_packet_header_t* header, const SP_packet_payload_t* payload, SP_callback_function_t function, SP_RUDP_transmission_result_t* result) {
	if((function != NULL) && (result == NULL)) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "RUDP: Trying to send packet with callback, but no memory for the results has been allocated.\n");
		return;
	} else if(header == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "RUDP: Trying to send packet with without header.\n");
		if(function != NULL) {
			result->errno = SP_ERROR_PARAMETER;
			function(result);
		}
		return;
	} else if((header->payload_length > 0) && (payload == NULL)) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "RUDP: Trying to send non-empty packet with without payload.\n");
		if(function != NULL) {
			result->errno = SP_ERROR_PARAMETER;
			function(result);
		}
		return;
	} else if(header->to == SP_LOCAL_ADDR) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "RUDP: Trying to send packet to local node.\n");
		if(function != NULL) {
			result->errno = SP_ERROR_PARAMETER;
			function(result);
		}
		return;
	} else if(header->to == SP_BROADCAST_ADDR) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "RUDP: Broadcasts are not supported, ignoring packet...\n");
		if(function != NULL) {
			result->errno = SP_ERROR_PARAMETER;
			function(result);
		}
		return;
	}

	SP_RUDP_total_transmissions++;

	if(list_size(SP_RUDP_transmission_queue) >= RUDP_MAX_QUEUED_PACKETS) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "RUDP: Queue full, dropping packet...\n");
		if(function != NULL) {
			result->errno = SP_ERROR_BUSY;
			function(result);
		}
		SP_RUDP_failed_transmissions++;
		return;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "RUDP: Reliably sending packet with header at 0x%X and payload at 0x%X...\n", header, payload);
	SP_RUDP_transmission_t* transmission;
	SP_MALLOC(transmission, SP_RUDP_transmission_t);
	SP_MALLOC(transmission->packet.header, SP_packet_header_t);
	memcpy(transmission->packet.header, header, sizeof(SP_packet_header_t));
	if(header->payload_length > 0) {
		char* p;
		SP_MALLOC_ARRAY(p, char, header->payload_length);
		memcpy(p, payload, header->payload_length);
		transmission->packet.payload = (SP_packet_payload_t*) p;
	} else {
		transmission->packet.payload = NULL;
	}
	transmission->retries = RUDP_PACKET_RETRIES;
	transmission->acked = false;
	if(function == NULL) {
		transmission->function = NULL;
		transmission->result = NULL;
	} else {
		transmission->function = function;
		transmission->result = result;
		transmission->result->errno = SP_ERROR_NOT_SET;
	}

	list_add(SP_RUDP_transmission_queue, transmission);
	send_RUDP_DATA(transmission->packet.header, transmission->packet.payload);
	callback_add(RUDP_send_timeout, transmission, SP_EXPECTED_MULTI_HOP_RTT);
}

/**
 * @brief Return the MTU size supported by RUDP.
 *
 * @return Maximal Transmission Unit (MTU) size in bytes.
 */
size_t RUDP_max_payload_size() {
	return SP_max_payload_size() - sizeof(SP_RUDP_DATA_packet_t);
}



/*
 * Interface provided by this module for direct commands.
 */

/**
 * @brief RUDP command function.
 *
 * @param argc Number of command parameters.
 * @param argv Array of strings used as command parameters.
 * @return <code>true</code> if the command was executed successfully,
 * <code>false</code> otherwise.
 */
bool RUDP_command(int argc, const char*const* argv) {
	return false;
}
