/**
 * @file service.c
 *
 * @brief Service main file.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service.c 2548 2008-02-13 19:20:58Z wittenbu $
 */

#include <stdlib.h>
#include <string.h>

#include "../types.h"
#include "../vars.h"
#include "../interface.h"
#include "../interface_internal.h"
#include "../util/debug.h"
#include "../util/net.h"
#include "service.h"
#include "service_types.h"
#include "service_vars.h"
#include "service_client.h"
#include "service_server.h"
#include "service_management.h"
#include "service_replication.h"
#include "service_util.h"



/*
 * Interface to main module.
 */

/**
 * @brief Initialize the service module.
 */
void SERVICE_init() {
	// Client variables.
	list_alloc(&SP_SERVICE_service_cache);
	SP_SERVICE_total_data_requests = 0;
	SP_SERVICE_successful_data_requests = 0;
	SP_SERVICE_failed_data_requests = 0;
	SP_SERVICE_total_successful_data_request_delay = 0;
	SP_SERVICE_total_data_updates = 0;
	SP_SERVICE_successful_data_updates = 0;
	SP_SERVICE_failed_data_updates = 0;
	SP_SERVICE_total_stream_requests = 0;
	SP_SERVICE_successful_stream_requests = 0;
	SP_SERVICE_failed_stream_requests = 0;
	SP_SERVICE_expected_stream_packets = 0;
	SP_SERVICE_received_stream_packets = 0;

	// Server variables.
	list_alloc(&SP_SERVICE_hosted_services);
	list_alloc(&SP_SERVICE_served_streams);
	list_alloc(&SP_SERVICE_incoming_replications);
	list_alloc(&SP_SERVICE_outgoing_replications);
	SP_SERVICE_total_replications = 0;
	SP_SERVICE_successful_replications = 0;
	SP_SERVICE_failed_replications = 0;
	SP_SERVICE_total_migrations = 0;
	SP_SERVICE_successful_migrations = 0;
	SP_SERVICE_failed_migrations = 0;
}

/**
 * @brief Start the service module.
 */
void SERVICE_start() {
}

/**
 * @brief Stop the service module.
 */
void SERVICE_stop() {
}

/**
 * @brief Finalize the service module.
 */
void SERVICE_finalize() {
	// Client variables.
// NOTE: Data requests are not acknowledged, hence we can't tell whether they failed.
/*
	if(SP_SERVICE_total_data_requests != (SP_SERVICE_successful_data_requests + SP_SERVICE_failed_data_requests)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Data request counters don't add up (%d != (%d + %d)).\n", SP_SERVICE_total_data_requests, SP_SERVICE_successful_data_requests, SP_SERVICE_failed_data_requests);
		abort();
	}
*/
	if(SP_SERVICE_total_data_updates != (SP_SERVICE_successful_data_updates + SP_SERVICE_failed_data_updates)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Data update counters don't add up (%d != (%d + %d)).\n", SP_SERVICE_total_data_updates, SP_SERVICE_successful_data_updates, SP_SERVICE_failed_data_updates);
		abort();
	}
	if(SP_SERVICE_total_stream_requests != (SP_SERVICE_successful_stream_requests + SP_SERVICE_failed_stream_requests)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Stream request counters don't add up (%d != (%d + %d)).\n", SP_SERVICE_total_stream_requests, SP_SERVICE_successful_stream_requests, SP_SERVICE_failed_stream_requests);
		abort();
	}
	list_free(SP_SERVICE_service_cache);

	// Server variables.
	if(SP_SERVICE_total_replications != (SP_SERVICE_successful_replications + SP_SERVICE_failed_replications)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Replication counters don't add up (%d != (%d + %d)).\n", SP_SERVICE_total_replications, SP_SERVICE_successful_replications, SP_SERVICE_failed_replications);
		abort();
	}
	if(SP_SERVICE_total_migrations != (SP_SERVICE_successful_migrations + SP_SERVICE_failed_migrations)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Migration counters don't add up (%d != (%d + %d)).\n", SP_SERVICE_total_migrations, SP_SERVICE_successful_migrations, SP_SERVICE_failed_migrations);
		abort();
	}
	list_free(SP_SERVICE_hosted_services);
	list_free(SP_SERVICE_served_streams);
	list_free(SP_SERVICE_incoming_replications);
	list_free(SP_SERVICE_outgoing_replications);
}

/**
 * @brief Service packet classifier.
 *
 * @param header Pointer to header of an incoming packet.
 * @param payload Pointer to payload of an incoming packet.
 */
void SERVICE_recv(const SP_packet_header_t* header, const SP_packet_payload_t* payload) {
	if(header->app != SP_APP_SERVICE) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Received non-SERVICE packet.\n");
		abort();
	}

	switch(header->type) {
	case SP_TYPE_NONE:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Received packet with uninitialized type field from node %s.\n", addr_to_str(header->from));
		break;
	case SP_TYPE_SERVICE_CLIENT_DATA_REQUEST:
		if(header->payload_length != sizeof(SP_SERVICE_CLIENT_DATA_REQUEST_packet_t)) {
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Received CLIENT DATA REQUEST packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SP_SERVICE_CLIENT_DATA_REQUEST_packet_t));
			abort();
		}
		handle_SERVICE_CLIENT_DATA_REQUEST(header, (SP_SERVICE_CLIENT_DATA_REQUEST_packet_t*) payload);
		break;
	case SP_TYPE_SERVICE_CLIENT_DATA_UPDATE:
		handle_SERVICE_CLIENT_DATA_UPDATE(header, payload);
		break;
	case SP_TYPE_SERVICE_CLIENT_STREAM_REQUEST:
		if(header->payload_length != sizeof(SP_SERVICE_CLIENT_STREAM_REQUEST_packet_t)) {
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Received CLIENT STREAM REQUST packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SP_SERVICE_CLIENT_STREAM_REQUEST_packet_t));
			abort();
		}
		handle_SERVICE_CLIENT_STREAM_REQUEST(header, (SP_SERVICE_CLIENT_STREAM_REQUEST_packet_t*) payload);
		break;
	case SP_TYPE_SERVICE_SERVER_DATA_ITEM:
		handle_SERVICE_SERVER_DATA_ITEM(header, payload);
		break;
	case SP_TYPE_SERVICE_SERVER_STREAM_ITEM:
		handle_SERVICE_SERVER_STREAM_ITEM(header, payload);
		break;
	case SP_TYPE_SERVICE_SERVER_NOT_AVAILABLE:
		if(header->payload_length != sizeof(SP_SERVICE_SERVER_NOT_AVAILABLE_packet_t)) {
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Received SERVER NOT AVAILABLE packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SP_SERVICE_SERVER_NOT_AVAILABLE_packet_t));
			abort();
		}
		handle_SERVICE_SERVER_NOT_AVAILABLE(header, (SP_SERVICE_SERVER_NOT_AVAILABLE_packet_t*) payload);
		break;
	case SP_TYPE_SERVICE_REPLICATION_REQUEST:
		if(header->payload_length != sizeof(SP_SERVICE_REPLICATION_REQUEST_packet_t)) {
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Received REPLICATION REQUEST packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SP_SERVICE_REPLICATION_REQUEST_packet_t));
			abort();
		}
		handle_SERVICE_REPLICATION_REQUEST(header, (SP_SERVICE_REPLICATION_REQUEST_packet_t*) payload);
		break;
	case SP_TYPE_SERVICE_REPLICATION_DATA:
		handle_SERVICE_REPLICATION_DATA(header, payload);
		break;
	case SP_TYPE_SERVICE_REPLICATION_ACK:
		if(header->payload_length != sizeof(SP_SERVICE_REPLICATION_ACK_packet_t)) {
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Received REPLICATION ACK packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SP_SERVICE_REPLICATION_ACK_packet_t));
			abort();
		}
		handle_SERVICE_REPLICATION_ACK(header, (SP_SERVICE_REPLICATION_ACK_packet_t*) payload);
		break;
	case SP_TYPE_SERVICE_REPLICATION_NACK:
		if(header->payload_length != sizeof(SP_SERVICE_REPLICATION_NACK_packet_t)) {
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Received REPLICATION NACK packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SP_SERVICE_REPLICATION_NACK_packet_t));
			abort();
		}
		handle_SERVICE_REPLICATION_NACK(header, (SP_SERVICE_REPLICATION_NACK_packet_t*) payload);
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Received packet of unknown type 0x%X from node %s.\n", header->type, addr_to_str(header->from));
		abort();
	}
}



/*
 * Interface provided by this module.
 */

/**
 * @brief Request specific service data of a service.
 *
 * @param service_id ID of the service from which to request data.
 * @param offset Offset in service data.
 * @param size Size of requested service data.
 */
void SERVICE_request_data(const char* service_id, u_int16_t offset, size_t size) {
	if((strlen(service_id) + 1) > SP_SERVICE_ID_LENGTH) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Service ID \"%s\" is too long (%d + 1 > %d).\n", service_id, strlen(service_id), SP_SERVICE_ID_LENGTH);
		return;
	}

	request_data(service_id, offset, size);
}

/**
 * @brief Update specific service data of a service.
 *
 * @param service_id ID of the service of which to update data.
 * @param offset Offset in service data.
 * @param size Size of service data to update.
 * @param data Pointer to new service data.
 */
void SERVICE_update_data(const char* service_id, u_int16_t offset, size_t size, const char* data) {
	if((strlen(service_id) + 1) > SP_SERVICE_ID_LENGTH) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Service ID \"%s\" is too long (%d + 1 > %d).\n", service_id, strlen(service_id), SP_SERVICE_ID_LENGTH);
		return;
	} else if(data == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: No data provided.\n");
		return;
	}

	update_data(service_id, offset, size, data);
}

/**
 * @brief Request a stream of specific service data of a service.
 *
 * @param service_id ID of the service from which to request data.
 * @param offset Offset in service data.
 * @param size Size of requested service data.
 * @param interval Interval at which to send stream packets in seconds.
 * @param duration Time during which the stream is to be active in seconds.
 */
void SERVICE_request_stream(const char* service_id, u_int16_t offset, size_t size, SP_time_t interval, SP_time_t duration) {
	if((strlen(service_id) + 1) > SP_SERVICE_ID_LENGTH) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Service ID \"%s\" is too long (%d + 1 > %d).\n", service_id, strlen(service_id), SP_SERVICE_ID_LENGTH);
		return;
	} else if(interval < 0) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Interval has negative value (%f).\n", interval);
		return;
	} else if(duration < 0) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Duration has negative value (%f).\n", duration);
		return;
	}

	request_stream(service_id, offset, size, interval, duration);
}

/**
 * @brief Create a new service on this node.
 *
 * @param service_id ID of the new service.
 * @param service_data_size Size of the service data in bytes.
 */
void SERVICE_create_service(const char* service_id, size_t service_data_size) {
	if((strlen(service_id) + 1) > SP_SERVICE_ID_LENGTH) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: New service ID \"%s\" is too long (%d + 1 > %d).\n", service_id, strlen(service_id), SP_SERVICE_ID_LENGTH);
		return;
	}

	SP_SERVICE_service_t* service = alloc_service(service_id, service_data_size);
	if(service == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Can't create new service \"%s\".\n", service_id);
		return;
	}
#ifdef SP_ENABLE_DEBUG
	char* p;
	SP_MALLOC_ARRAY(p, char, service_data_size);
	memset(p, 0xDA, service_data_size);
	update_service_data(service, 0x00, service_data_size, p);	// Fill service data memory with a known pattern.
#endif	// SP_ENABLE_DEBUG
	start_service(service);
}

/**
 * @brief Shutdown a service on this node.
 *
 * @param service_id ID of the service to shutdown.
 */
void SERVICE_shutdown_service(const char* service_id) {
	if((strlen(service_id) + 1) > SP_SERVICE_ID_LENGTH) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Service ID \"%s\" is too long (%d + 1 > %d).\n", service_id, strlen(service_id), SP_SERVICE_ID_LENGTH);
		return;
	}

	SP_SERVICE_service_t* service = get_service(service_id);
	if(service == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Can't shutdown unknown service \"%s\".\n", service_id);
		return;
	}
	stop_service(service);
	free_service(service);
}

/**
 * @brief Retrieve a pointer to an active service on this node.
 *
 * @param service_id ID of the service to be retrieved.
 * @return Pointer to the service if it is active on this node, NULL otherwise.
 */
SP_SERVICE_service_t* SERVICE_get_active_service(const char* service_id) {
	if((strlen(service_id) + 1) > SP_SERVICE_ID_LENGTH) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Service ID \"%s\" is too long (%d + 1 > %d).\n", service_id, strlen(service_id), SP_SERVICE_ID_LENGTH);
		return NULL;
	}

	return get_active_service(service_id);
}

/**
 * @brief Replicate a service from this node to another node.
 *
 * @param service_id ID of the service to replicate.
 * @param to Address of the destination node.
 * @param callback Callback function.
 * @param result Pointer to replication result.
 */
void SERVICE_replicate_service(const char* service_id, SP_addr_t to, SP_callback_function_t callback, SP_SERVICE_replication_result_t* result) {
	if((strlen(service_id) + 1) > SP_SERVICE_ID_LENGTH) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Service ID \"%s\" is too long (%d + 1 > %d).\n", service_id, strlen(service_id), SP_SERVICE_ID_LENGTH);
		return;
	}
	SP_SERVICE_service_t* service = get_active_service(service_id);
	if(service == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Can't replicate service \"%s\", service not active.\n", service_id);
		return;
	}

	replicate_service(service, to, callback, result);
}

/**
 * @brief Migrate a service from this node to another node.
 *
 * @param service_id ID of the service to migrate.
 * @param to Address of the destination node.
 * @param callback Callback function.
 * @param result Pointer to migration result.
 */
void SERVICE_migrate_service(const char* service_id, SP_addr_t to, SP_callback_function_t callback, SP_SERVICE_migration_result_t* result) {
	if((strlen(service_id) + 1) > SP_SERVICE_ID_LENGTH) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Service ID \"%s\" is too long (%d + 1 > %d).\n", service_id, strlen(service_id), SP_SERVICE_ID_LENGTH);
		return;
	}
	SP_SERVICE_service_t* service = get_active_service(service_id);
	if(service == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Can't migrate service \"%s\", service not active.\n", service_id);
		return;
	}

	migrate_service(service, to, callback, result);
}

/**
 * @brief Check whether two service IDs match.
 *
 * @param service_id_1 First service ID.
 * @param service_id_2 Seconds service ID.
 * @return <code>true</code> if the two service IDs are equal,
 * <code>false</code> otherwise.
 */
bool SERVICE_same_service_id(const char* service_id_1, const char* service_id_2) {
	return same_service_id(service_id_1, service_id_2);
}



/*
 * Interface provided by this module for direct commands.
 */

/**
 * @brief Service command replication callback function.
 *
 * @param parameter Replication result.
 */
static void SERVICE_command_replicate_service_callback(SP_callback_parameter_t parameter) {
	SP_SERVICE_replication_result_t* result = (SP_SERVICE_replication_result_t*) parameter;
	SP_SERVICE_service_t* service = (SP_SERVICE_service_t*) result->parameter;

	switch(result->errno) {
	case SP_ERROR_OK:
		SP_DEBUG(SP_DEBUG_LEVEL_NOTE, "SERVICE: Replication of service \"%s\" was successful.\n", service->id);
		break;
	case SP_ERROR_PARAMETER:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Replication of service \"%s\" failed due to an invalid parameter.\n", service->id);
		break;
	case SP_ERROR_TIMEOUT:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Replication of service \"%s\" timed out.\n", service->id);
		break;
	case SP_ERROR_BUSY:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Replication of service \"%s\" failed because system is busy.\n", service->id);
		break;
	case SP_ERROR_IN_PROGRESS:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Replication of service \"%s\" failed due to an invalid parameter.\n", service->id);
		break;
	case SP_ERROR_COMM_ERROR:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Replication of service \"%s\" failed due to communication errors.\n", service->id);
		break;
	case SP_ERROR_REMOTE_BUSY:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Replication of service \"%s\" failed because the remote system refused.\n", service->id);
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Replication callback with unknown error code 0x%X.\n", result->errno);
		abort();
	}
	SP_FREE(result);
}

/**
 * @brief Service command migration callback function.
 *
 * @param parameter Migration result.
 */
static void SERVICE_command_migrate_service_callback(SP_callback_parameter_t parameter) {
	SP_SERVICE_migration_result_t* result = (SP_SERVICE_migration_result_t*) parameter;

	switch(result->errno) {
	case SP_ERROR_OK:
		SP_DEBUG(SP_DEBUG_LEVEL_NOTE, "SERVICE: Migration of service \"%s\" was successful.\n", result->service_id);
		break;
	case SP_ERROR_PARAMETER:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Migration of service \"%s\" failed due to an invalid parameter.\n", result->service_id);
		break;
	case SP_ERROR_TIMEOUT:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Migration of service \"%s\" timed out.\n", result->service_id);
		break;
	case SP_ERROR_BUSY:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Migration of service \"%s\" failed because system is busy.\n", result->service_id);
		break;
	case SP_ERROR_IN_PROGRESS:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Migration of service \"%s\" failed due to an invalid parameter.\n", result->service_id);
		break;
	case SP_ERROR_COMM_ERROR:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Migration of service \"%s\" failed due to communication errors.\n", result->service_id);
		break;
	case SP_ERROR_REMOTE_BUSY:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Migration of service \"%s\" failed because the remote system refused.\n", result->service_id);
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Migration callback with unknown error code 0x%X.\n", result->errno);
		abort();
	}
	SP_FREE(result);
}

/**
 * @brief Service command function.
 *
 * @param argc Number of command parameters.
 * @param argv Array of strings used as command parameters.
 * @return <code>true</code> if the command was executed successfully,
 * <code>false</code> otherwise.
 */
bool SERVICE_command(int argc, const char*const* argv) {
	if(argc == 3) {
		if(strcmp(argv[1], "SERVICE_shutdown_service") == 0) {
			SERVICE_shutdown_service(argv[2]);
			return true;
		} else if(strcmp(argv[1], "SERVICE_is_local_service") == 0) {
			if(SERVICE_get_active_service(argv[2]) != NULL) {
				SP_DEBUG(SP_DEBUG_LEVEL_NOTE, "SERVICE: Service \"%s\" is hosted on this node.\n", argv[2]);
			} else {
				SP_DEBUG(SP_DEBUG_LEVEL_NOTE, "SERVICE: Service \"%s\" is not hosted on this node.\n", argv[2]);
			}
			return true;
		}
	} else if(argc == 4) {
		if(strcmp(argv[1], "SERVICE_create_service") == 0) {
			SERVICE_create_service(argv[2], atoi(argv[3]));
			return true;
		} else if(strcmp(argv[1], "SERVICE_replicate_service") == 0) {
			SP_SERVICE_service_t* service = get_active_service(argv[2]);
			if(service == NULL) {
				SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Can't replicate service \"%s\", service not active.\n", argv[2]);
				return false;
			}
			SP_SERVICE_replication_result_t* result;
			SP_MALLOC(result, SP_SERVICE_replication_result_t);
			result->parameter = service;
			SERVICE_replicate_service(argv[2], atoi(argv[3]), SERVICE_command_replicate_service_callback, result);
			return true;
		} else if(strcmp(argv[1], "SERVICE_migrate_service") == 0) {
			SP_SERVICE_service_t* service = get_active_service(argv[2]);
			if(service == NULL) {
				SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Can't migrate service \"%s\", service not active.\n", argv[2]);
				return false;
			}
			SP_SERVICE_migration_result_t* result;
			SP_MALLOC(result, SP_SERVICE_migration_result_t);
			result->parameter = service;
			SERVICE_migrate_service(argv[2], atoi(argv[3]), SERVICE_command_migrate_service_callback, result);
			return true;
		}
	} else if(argc == 5) {
		if(strcmp(argv[1], "SERVICE_request_data") == 0) {
			SERVICE_request_data(argv[2], atoi(argv[3]), atoi(argv[4]));
			return true;
		} else if(strcmp(argv[1], "SERVICE_update_data") == 0) {
			size_t size = atoi(argv[4]);
			char* data;
			SP_MALLOC_ARRAY(data, char, size);
			memset(data, 0xDB, size);	// Some byte pattern.
			SERVICE_update_data(argv[2], atoi(argv[3]), size, data);
			return true;
		}
	} else if(argc == 7) {
		if(strcmp(argv[1], "SERVICE_request_stream") == 0) {
			SERVICE_request_stream(argv[2], atoi(argv[3]), atoi(argv[4]), atof(argv[5]), atof(argv[6]));
			return true;
		}
	}
	return false;
}
