/**
 * @file service_client.c
 *
 * @brief Service - Client.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service_client.c 2585 2008-03-05 13:28:49Z wittenbu $
 */

#include <math.h>
#include <string.h>

#include "../types.h"
#include "../vars.h"
#include "../interface.h"
#include "../interface_internal.h"
#include "../util/debug.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "../rudp-light/rudp-light.h"
#include "../slp-light/slp-light.h"
#include "service.h"
#include "service_types.h"
#include "service_vars.h"
#include "service_client.h"
#include "service_management.h"
#include "service_util.h"



/**
 * @brief Number of service instances to locate before selecting the best
 * server.
 */
#define CLIENT_MIN_SERVICE_INSTANCES	1



/*
 * Packet handling functions.
 */

/**
 * @brief Send SERVICE DATA REQUEST packet.
 *
 * @param to Address of destination node.
 * @param service_id ID of service of which data is requested.
 * @param offset Offset in service data.
 * @param size Request size in bytes.
 */
static void send_SERVICE_CLIENT_DATA_REQUEST(SP_addr_t to, const char* service_id, u_int16_t offset, size_t size) {
	SP_packet_header_t header;
	SP_SERVICE_CLIENT_DATA_REQUEST_packet_t request;

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Sending CLIENT DATA REQUEST for %d byte(s) at 0x%X of service \"%s\" to node %s...\n", size, offset, service_id, addr_to_str(to));
	init_header(&header);
	header.to = to;
	header.app = SP_APP_SERVICE;
	header.type = SP_TYPE_SERVICE_CLIENT_DATA_REQUEST;
	header.payload_length = sizeof(SP_SERVICE_CLIENT_DATA_REQUEST_packet_t);
	strncpy(request.service_id, service_id, SP_SERVICE_ID_LENGTH);
	request.offset = offset;
	request.size = size;
	request.start_time = SP_time();
	SP_send(&header, (SP_packet_payload_t*) &request);

	cache_increment_age(service_id);
}

/**
 * @brief RUDP callback for SERVICE DATA UPDATE packets.
 *
 * @param parameter RUDP transmission result.
 */
static void send_SERVICE_CLIENT_DATA_UPDATE_callback(SP_callback_parameter_t parameter) {
	SP_RUDP_transmission_result_t* result = (SP_RUDP_transmission_result_t*) parameter;

	if(result->errno == SP_ERROR_OK) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: CLIENT DATA UPDATE was successful.\n");
		SP_SERVICE_successful_data_updates++;
	} else {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: CLIENT DATA UPDATE failed (%d).\n", result->errno);
		SP_SERVICE_failed_data_updates++;
	}

	SP_FREE(result);
}

/**
 * @brief Send SERVICE DATA UPDATE packet.
 *
 * @param to Address of destination node.
 * @param service_id ID of service of which data is to be updated.
 * @param offset Offset in service data.
 * @param size Update size in bytes.
 * @param data Pointer to new service data.
 */
static void send_SERVICE_CLIENT_DATA_UPDATE(SP_addr_t to, const char* service_id, u_int16_t offset, size_t size, const char* data) {
	SP_packet_header_t header;
	SP_SERVICE_CLIENT_DATA_UPDATE_packet_t update;

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Sending CLIENT DATA UPDATE for %d byte(s) at 0x%X of service \"%s\" (0x%X...) to node %s...\n", size, offset, service_id, (u_int8_t) *data, addr_to_str(to));
	init_header(&header);
	header.to = to;
	header.app = SP_APP_SERVICE;
	header.type = SP_TYPE_SERVICE_CLIENT_DATA_UPDATE;
	header.payload_length = sizeof(SP_SERVICE_CLIENT_DATA_UPDATE_packet_t) + size;
	strncpy(update.service_id, service_id, SP_SERVICE_ID_LENGTH);
	update.offset = offset;

	SP_packet_payload_t* payload;
	SP_MALLOC_JOIN(payload, SP_packet_payload_t, &update, sizeof(SP_SERVICE_CLIENT_DATA_UPDATE_packet_t), data, size);
	SP_RUDP_transmission_result_t* result;
	SP_MALLOC(result, SP_RUDP_transmission_result_t);
	RUDP_send(&header, payload, send_SERVICE_CLIENT_DATA_UPDATE_callback, result);
	SP_FREE(payload);
}

/**
 * @brief RUDP callback for SERVICE STREAM REQUEST packets.
 *
 * @param parameter RUDP transmission result.
 */
static void send_SERVICE_CLIENT_STREAM_REQUEST_callback(SP_callback_parameter_t parameter) {
	SP_RUDP_transmission_result_t* result = (SP_RUDP_transmission_result_t*) parameter;

	if(result->errno == SP_ERROR_OK) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: CLIENT STREAM REQUEST was successful.\n");
		SP_SERVICE_successful_stream_requests++;
	} else {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: CLIENT STREAM REQUEST failed (%d).\n", result->errno);
		SP_SERVICE_failed_stream_requests++;
	}

	SP_FREE(result);
}

/**
 * @brief Send SERVICE STREAM REQUEST packet.
 *
 * @param to Address of destination node.
 * @param service_id ID of service of which data is requested.
 * @param offset Offset in service data.
 * @param size Request size in bytes.
 * @param interval Interval between streamed packets in seconds.
 * @param duration Time the stream is to be active in seconds.
 */
static void send_SERVICE_CLIENT_STREAM_REQUEST(SP_addr_t to, const char* service_id, u_int16_t offset, size_t size, SP_time_t interval, SP_time_t duration) {
	SP_packet_header_t header;
	SP_SERVICE_CLIENT_STREAM_REQUEST_packet_t request;

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Sending CLIENT STREAM REQUEST for %d byte(s) at 0x%X of service \"%s\" at interval %.3fs for a duration of %.2fs to node %s...\n", size, offset, service_id, interval, duration, addr_to_str(to));
	init_header(&header);
	header.to = to;
	header.app = SP_APP_SERVICE;
	header.type = SP_TYPE_SERVICE_CLIENT_STREAM_REQUEST;
	header.payload_length = sizeof(SP_SERVICE_CLIENT_STREAM_REQUEST_packet_t);
	strncpy(request.service_id, service_id, SP_SERVICE_ID_LENGTH);
	request.offset = offset;
	request.size = size;
	request.interval = interval;
	request.duration = duration;
	SP_RUDP_transmission_result_t* result;
	SP_MALLOC(result, SP_RUDP_transmission_result_t);
	RUDP_send(&header, (SP_packet_payload_t*) &request, send_SERVICE_CLIENT_STREAM_REQUEST_callback, result);

	cache_increment_age(service_id);
}

/**
 * @brief Handle incoming SERVICE DATA ITEM packet.
 *
 * Possibly redirect future requests to this service if server provided
 * corresponding information
 *
 * @param header Header of the incoming packet.
 * @param payload Payload of incoming packet containing the SERVICE DATA ITEM
 * header and the requested data itself.
 */
void handle_SERVICE_SERVER_DATA_ITEM(const SP_packet_header_t* header, const SP_packet_payload_t* payload) {
	SP_SERVICE_SERVER_DATA_ITEM_packet_t data_item;
	char* data;

	if(header->payload_length < sizeof(SP_SERVICE_SERVER_DATA_ITEM_packet_t)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Received CLIENT DATA ITEM packet with wrong payload length (%d < %d).\n", header->payload_length, sizeof(SP_SERVICE_SERVER_DATA_ITEM_packet_t));
		abort();
	}

	size_t data_size = header->payload_length - sizeof(SP_SERVICE_SERVER_DATA_ITEM_packet_t);
	SP_MALLOC_ARRAY(data, char, data_size);
	SP_SPLIT(&data_item, sizeof(SP_SERVICE_SERVER_DATA_ITEM_packet_t), data, data_size, payload);
	SP_time_t delay = SP_time() - data_item.start_time;
	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Received %d byte(s) of SERVER DATA ITEM for service \"%s\" (0x%X...) from node %s after %.2fs.\n", data_size, data_item.service_id, (u_int8_t) *data, addr_to_str(header->from), delay);
	SP_FREE(data);

	if(data_item.redirect_addr == SP_BROADCAST_ADDR) {
		cache_set(data_item.service_id, header->from);
	} else {
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Redirecting future request for service \"%s\" to node %s.\n", data_item.service_id, addr_to_str(data_item.redirect_addr));
		cache_set(data_item.service_id, data_item.redirect_addr);
	}

	SP_SERVICE_successful_data_requests++;
	SP_SERVICE_total_successful_data_request_delay += delay;
}

/**
 * @brief Handle incoming SERVICE STREAM ITEM packet.
 *
 * Possibly redirect future requests to this service if server provided
 * corresponding information
 *
 * @param header Header of the incoming packet.
 * @param payload Payload of incoming packet containing the SERVICE STREAM ITEM
 * header and the streamed data itself.
 */
void handle_SERVICE_SERVER_STREAM_ITEM(const SP_packet_header_t* header, const SP_packet_payload_t* payload) {
	SP_SERVICE_SERVER_STREAM_ITEM_packet_t stream_item;
	char* data;

	if(header->payload_length < sizeof(SP_SERVICE_SERVER_STREAM_ITEM_packet_t)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Received CLIENT STREAM ITEM packet with wrong payload length (%d < %d).\n", header->payload_length, sizeof(SP_SERVICE_SERVER_STREAM_ITEM_packet_t));
		abort();
	}

	size_t data_size = header->payload_length - sizeof(SP_SERVICE_SERVER_STREAM_ITEM_packet_t);
	SP_MALLOC_ARRAY(data, char, data_size);
	SP_SPLIT(&stream_item, sizeof(SP_SERVICE_SERVER_STREAM_ITEM_packet_t), data, data_size, payload);
	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Received %d byte(s) of SERVER STREAM ITEM (0x%X...) from node %s.\n", header->payload_length, (u_int8_t) *payload, addr_to_str(header->from));
	SP_FREE(data);

	if(stream_item.redirect_addr == SP_BROADCAST_ADDR) {
		cache_set(stream_item.service_id, header->from);
	} else {
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Redirecting future request to service \"%s\" to node %s.\n", stream_item.service_id, addr_to_str(stream_item.redirect_addr));
		cache_set(stream_item.service_id, stream_item.redirect_addr);
	}

	SP_SERVICE_received_stream_packets++;
}

/**
 * @brief Handle incoming SERVICE NOT AVAILABLE packet.
 *
 * Possibly redirect future requests to this service if server provided
 * corresponding information
 *
 * @param header Header of incoming packet.
 * @param not_available Data structure with redirection information.
 */
void handle_SERVICE_SERVER_NOT_AVAILABLE(const SP_packet_header_t* header, const SP_SERVICE_SERVER_NOT_AVAILABLE_packet_t* not_available) {
	SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Received of SERVER NOT AVAILABLE for service \"%s\" from node %s.\n", not_available->service_id, addr_to_str(header->from));

	if(not_available->redirect_addr != SP_BROADCAST_ADDR) {
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Redirecting future requests to service \"%s\" to node %s.\n", not_available->service_id, addr_to_str(not_available->redirect_addr));
		cache_set(not_available->service_id, not_available->redirect_addr);
	} else {
		SP_addr_t server;
		switch(cache_lookup(not_available->service_id, &server)) {
		case SP_ERROR_OK:
			if(server == header->from) {
				SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Removing cache entry for service \"%s\" pointing to node %s.\n", not_available->service_id, addr_to_str(server));
				cache_remove(not_available->service_id);
			}
			break;
		case SP_ERROR_NOT_FOUND:
		case SP_ERROR_TIMEOUT:
			break;
		default:
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Cache lookup returned invalid error code.\n");
			abort();
		}
	}
}



/*
 * Control function.
 */

/**
 * @brief Callback for retried data request if SLP query was in progress.
 *
 * Proceed with data request if query was successful, abort otherwise
 *
 * @param parameter SLP query result.
 */
static void request_data_retry(SP_callback_parameter_t parameter) {
	SP_SLP_query_result_t* slp_query_result = (SP_SLP_query_result_t*) parameter;
	SP_SERVICE_data_request_t* request = (SP_SERVICE_data_request_t*) slp_query_result->parameter;

	SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Retrying request for %d byte(s) of service \"%s\"...\n", request->size, request->service_id);
	SP_addr_t server;
	switch(cache_lookup(request->service_id, &server)) {
	case SP_ERROR_OK:
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Reusing cached server for service \"%s\"...\n", request->service_id);
		send_SERVICE_CLIENT_DATA_REQUEST(server, request->service_id, request->offset, request->size);
		break;
	case SP_ERROR_NOT_FOUND:
	case SP_ERROR_TIMEOUT:
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Service \"%s\" is not reachable, dropping data request...\n", request->service_id);
		SP_SERVICE_failed_data_requests++;
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Cache lookup returned invalid error code.\n");
		abort();
	}
	SP_FREE(request);
	list_free(slp_query_result->service_instances);
	SP_FREE(slp_query_result);
}

/**
 * @brief Data request callback function for SLP queries.
 *
 * Proceed with data request if query was succesful, schedule retry if query is
 * in progress, abort otherwise.
 *
 * @param parameter SLP query result
 */
static void request_data_callback(SP_callback_parameter_t parameter) {
	SP_addr_t server;
	SP_SLP_query_result_t* slp_query_result = (SP_SLP_query_result_t*) parameter;
	SP_SERVICE_data_request_t* request = (SP_SERVICE_data_request_t*) slp_query_result->parameter;

	switch(slp_query_result->errno) {
	case SP_ERROR_OK:
		server = selectBestServer(slp_query_result->service_instances);
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Using node %s as server for service \"%s\"...\n", addr_to_str(server), request->service_id);
		cache_set(request->service_id, server);
		send_SERVICE_CLIENT_DATA_REQUEST(server, request->service_id, request->offset, request->size);
		break;
	case SP_ERROR_TIMEOUT:
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Service \"%s\" is not reachable, dropping data request...\n", request->service_id);
		cache_tag_not_reachable(request->service_id);
		SP_SERVICE_failed_data_requests++;
		break;
	case SP_ERROR_IN_PROGRESS:
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Query for service \"%s\" is already in progress, scheduling retry in %.2fs...\n", request->service_id, SP_TIMEOUT);
		callback_add(request_data_retry, slp_query_result, SP_TIMEOUT);
		return;	// Don't free the resources at the end of this function.
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: SERVICE_request_data() callback with error code %d.\n", slp_query_result->errno);
		abort();
	}

	SP_FREE(request);
	list_free(slp_query_result->service_instances);
	SP_FREE(slp_query_result);
}

/**
 * @brief Request specific service data of a service.
 *
 * Perform SLP query for service in case it is not contained in the service
 * cache.
 *
 * @param service_id ID of the service from which to request data.
 * @param offset Offset in service data.
 * @param size Size of requested service data.
 */
void request_data(const char* service_id, u_int16_t offset, size_t size) {
	SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Requesting %d byte(s) at 0x%X of service \"%s\"...\n", size, offset, service_id);
	SP_SERVICE_total_data_requests++;

	SP_addr_t server;
	switch(cache_lookup(service_id, &server)) {
	case SP_ERROR_OK:
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Reusing cached server for service \"%s\"...\n", service_id);
		send_SERVICE_CLIENT_DATA_REQUEST(server, service_id, offset, size);
		break;
	case SP_ERROR_NOT_FOUND:
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: No server in cache for service \"%s\", starting SLP query...\n", service_id);
		SP_SLP_query_result_t* slp_query_result;
		SP_MALLOC(slp_query_result, SP_SLP_query_result_t);
		list_alloc(&slp_query_result->service_instances);
		SP_SERVICE_data_request_t* request;
		SP_MALLOC(request, SP_SERVICE_data_request_t);
		strncpy(request->service_id, service_id, SP_SERVICE_ID_LENGTH);
		request->offset = offset;
		request->size = size;
		slp_query_result->parameter = request;
		SLP_locate_service(service_id, CLIENT_MIN_SERVICE_INSTANCES, request_data_callback, slp_query_result);
		break;
	case SP_ERROR_TIMEOUT:
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Service \"%s\" is not reachable, dropping data request...\n", service_id);
		SP_SERVICE_failed_data_requests++;
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Cache lookup returned invalid error code.\n");
		abort();
	}
}

/**
 * @brief Callback for retried data update if SLP query was in progress.
 *
 * Proceed with data update if query was successful, abort otherwise
 *
 * @param parameter SLP query result.
 */
static void update_data_retry(SP_callback_parameter_t parameter) {
	SP_SLP_query_result_t* slp_query_result = (SP_SLP_query_result_t*) parameter;
	SP_SERVICE_data_update_t* update = (SP_SERVICE_data_update_t*) slp_query_result->parameter;

	SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Retrying update of %d byte(s) of service \"%s\"...\n", update->size, update->service_id);
	SP_addr_t server;
	switch(cache_lookup(update->service_id, &server)) {
	case SP_ERROR_OK:
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Reusing cached server for service \"%s\"...\n", update->service_id);
		send_SERVICE_CLIENT_DATA_UPDATE(server, update->service_id, update->offset, update->size, update->data);
		break;
	case SP_ERROR_NOT_FOUND:
	case SP_ERROR_TIMEOUT:
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Service \"%s\" is not reachable, dropping data update...\n", update->service_id);
		SP_SERVICE_failed_data_updates++;
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Cache lookup returned invalid error code.\n");
		abort();
	}
	SP_FREE(update);
	list_free(slp_query_result->service_instances);
	SP_FREE(slp_query_result);
}

/**
 * @brief Data update callback function for SLP queries.
 *
 * Proceed with data update if query was succesful, schedule retry if query is
 * in progress, abort otherwise.
 *
 * @param parameter SLP query result
 */
static void update_data_callback(SP_callback_parameter_t parameter) {
	SP_addr_t server;
	SP_SLP_query_result_t* slp_query_result = (SP_SLP_query_result_t*) parameter;
	SP_SERVICE_data_update_t* update = (SP_SERVICE_data_update_t*) slp_query_result->parameter;

	switch(slp_query_result->errno) {
	case SP_ERROR_OK:
		server = selectBestServer(slp_query_result->service_instances);
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Using node %s as server for service \"%s\"...\n", addr_to_str(server), update->service_id);
		cache_set(update->service_id, server);
		send_SERVICE_CLIENT_DATA_UPDATE(server, update->service_id, update->offset, update->size, update->data);
		break;
	case SP_ERROR_TIMEOUT:
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Service \"%s\" is not reachable, dropping data update...\n", update->service_id);
		cache_tag_not_reachable(update->service_id);
		SP_SERVICE_failed_data_updates++;
		break;
	case SP_ERROR_IN_PROGRESS:
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Query for service \"%s\" is already in progress, scheduling retry in %.2fs...\n", update->service_id, SP_TIMEOUT);
		callback_add(update_data_retry, slp_query_result, SP_TIMEOUT);
		return;	// Don't free the resources at the end of this function.
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: SERVICE_request_data() callback with error code %d.\n", slp_query_result->errno);
		abort();
	}

	SP_FREE(update);
	list_free(slp_query_result->service_instances);
	SP_FREE(slp_query_result);
}

/**
 * @brief Update specific service data of a service.
 *
 * Perform SLP query for service in case it is not contained in the service
 * cache.
 *
 * @param service_id ID of the service of which to update data.
 * @param offset Offset in service data.
 * @param size Size of service data to update.
 * @param data Pointer to new service data.
 */
void update_data(const char* service_id, u_int16_t offset, size_t size, const char* data) {
	SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Updating %d byte(s) at 0x%X of service \"%s\"...\n", size, offset, service_id);
	SP_SERVICE_total_data_updates++;

	SP_addr_t server;
	switch(cache_lookup(service_id, &server)) {
	case SP_ERROR_OK:
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Reusing cached server for service \"%s\"...\n", service_id);
		send_SERVICE_CLIENT_DATA_UPDATE(server, service_id, offset, size, data);
		break;
	case SP_ERROR_NOT_FOUND:
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: No server in cache for service \"%s\", starting SLP query...\n", service_id);
		SP_SLP_query_result_t* slp_query_result;
		SP_MALLOC(slp_query_result, SP_SLP_query_result_t);
		list_alloc(&slp_query_result->service_instances);
		SP_SERVICE_data_update_t* update;
		SP_MALLOC(update, SP_SERVICE_data_update_t);
		strncpy(update->service_id, service_id, SP_SERVICE_ID_LENGTH);
		update->offset = offset;
		update->size = size;
		update->data = data;
		slp_query_result->parameter = update;
		SLP_locate_service(service_id, CLIENT_MIN_SERVICE_INSTANCES, update_data_callback, slp_query_result);
		break;
	case SP_ERROR_TIMEOUT:
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Service \"%s\" is not reachable, dropping data update...\n", service_id);
		SP_SERVICE_failed_data_updates++;
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Cache lookup returned invalid error code.\n");
		abort();
	}
}

/**
 * @brief Callback for retried stream request if SLP query was in progress.
 *
 * Proceed with stream request if query was successful, abort otherwise
 *
 * @param parameter SLP query result.
 */
static void request_stream_retry(SP_callback_parameter_t parameter) {
	SP_SLP_query_result_t* slp_query_result = (SP_SLP_query_result_t*) parameter;
	SP_SERVICE_stream_request_t* request = (SP_SERVICE_stream_request_t*) slp_query_result->parameter;

	SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Retrying request for %d byte(s) of service \"%s\" at interval %.2fs for a duration of %.2fs...\n", request->size, request->service_id, request->interval, request->duration);
	SP_addr_t server;
	switch(cache_lookup(request->service_id, &server)) {
	case SP_ERROR_OK:
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Reusing cached server for service \"%s\"...\n", request->service_id);
		send_SERVICE_CLIENT_STREAM_REQUEST(server, request->service_id, request->offset, request->size, request->interval, request->duration);
		break;
	case SP_ERROR_NOT_FOUND:
	case SP_ERROR_TIMEOUT:
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Service \"%s\" is not reachable, dropping stream request...\n", request->service_id);
		SP_SERVICE_failed_stream_requests++;
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Cache lookup returned invalid error code.\n");
		abort();
	}
	SP_FREE(request);
	list_free(slp_query_result->service_instances);
	SP_FREE(slp_query_result);
}

/**
 * @brief Stream request callback function for SLP queries.
 *
 * Proceed with stream request if query was succesful, schedule retry if query
 * is in progress, abort otherwise.
 *
 * @param parameter SLP query result
 */
static void request_stream_callback(SP_callback_parameter_t parameter) {
	SP_addr_t server;
	SP_SLP_query_result_t* slp_query_result = (SP_SLP_query_result_t*) parameter;
	SP_SERVICE_stream_request_t* request = (SP_SERVICE_stream_request_t*) slp_query_result->parameter;

	switch(slp_query_result->errno) {
	case SP_ERROR_OK:
		server = selectBestServer(slp_query_result->service_instances);
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Using node %s as server for service \"%s\"...\n", addr_to_str(server), request->service_id);
		cache_set(request->service_id, server);
		send_SERVICE_CLIENT_STREAM_REQUEST(server, request->service_id, request->offset, request->size, request->interval, request->duration);
		break;
	case SP_ERROR_TIMEOUT:
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Service \"%s\" is not reachable, dropping stream request...\n", request->service_id);
		cache_tag_not_reachable(request->service_id);
		SP_SERVICE_failed_stream_requests++;
		break;
	case SP_ERROR_IN_PROGRESS:
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Query for service \"%s\" is already in progress, scheduling retry in %.2fs...\n", request->service_id, SP_TIMEOUT);
		callback_add(request_stream_retry, slp_query_result, SP_TIMEOUT);
		return;	// Don't free the resources at the end of this function.
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: SERVICE_request_data() callback with error code %d.\n", slp_query_result->errno);
		abort();
	}

	SP_FREE(request);
	list_free(slp_query_result->service_instances);
	SP_FREE(slp_query_result);
}

/**
 * @brief Request a stream of specific service data of a service.
 *
 * Perform SLP query for service in case it is not contained in the service
 * cache.
 *
 * @param service_id ID of the service from which to request data.
 * @param offset Offset in service data.
 * @param size Size of requested service data.
 * @param interval Interval at which to send stream packets in seconds.
 * @param duration Time during which the stream is to be active in seconds.
 */
void request_stream(const char* service_id, u_int16_t offset, size_t size, SP_time_t interval, SP_time_t duration) {
	SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Requesting a stream of %d byte(s) at 0x%X of service \"%s\" at interval %.2fs for a duration of %.2fs...\n", size, offset, service_id, interval, duration);
	SP_SERVICE_total_stream_requests++;

	// NOTE: With the current implementation, we cannot tell different
	// streams apart. As far as counting packets is concerned, this
	// shouldn't be a problem because RUDP allows us to tell apart whether
	// we have lost single packets or a stream request was never received.
	SP_SERVICE_expected_stream_packets = ((u_int32_t) (duration / interval)) + 1;

	SP_addr_t server;
	switch(cache_lookup(service_id, &server)) {
	case SP_ERROR_OK:
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Reusing cached server for service \"%s\"...\n", service_id);
		send_SERVICE_CLIENT_STREAM_REQUEST(server, service_id, offset, size, interval, duration);
		break;
	case SP_ERROR_NOT_FOUND:
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: No server in cache for service \"%s\", starting SLP query...\n", service_id);
		SP_SLP_query_result_t* slp_query_result;
		SP_MALLOC(slp_query_result, SP_SLP_query_result_t);
		list_alloc(&slp_query_result->service_instances);
		SP_SERVICE_stream_request_t* request;
		SP_MALLOC(request, SP_SERVICE_stream_request_t);
		strncpy(request->service_id, service_id, SP_SERVICE_ID_LENGTH);
		request->offset = offset;
		request->size = size;
		request->interval = interval;
		request->duration = duration;
		slp_query_result->parameter = request;
		SLP_locate_service(service_id, CLIENT_MIN_SERVICE_INSTANCES, request_stream_callback, slp_query_result);
		break;
	case SP_ERROR_TIMEOUT:
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Service \"%s\" is not reachable, dropping stream request...\n", service_id);
		SP_SERVICE_failed_stream_requests++;
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Cache lookup returned invalid error code.\n");
		abort();
	}
}
