/**
 * @file service_types.h
 *
 * @brief Service (types).
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service_types.h 2585 2008-03-05 13:28:49Z wittenbu $
 */

#ifndef SERVICE_TYPES_H
#define SERVICE_TYPES_H

#include "../types.h"
#include "../placement/placement_types.h"

/**
 * @brief Length of the service ID string.
 */
#define SP_SERVICE_ID_LENGTH	8

#define SP_SERVICE_STATE_INIT		0x0	///< Service is initializing.
#define SP_SERVICE_STATE_RUNNING	0x1	///< Service is running.
#define SP_SERVICE_STATE_REPLICATING	0x2	///< Service is replicating to another node.
#define SP_SERVICE_STATE_REDIRECTING	0x3	///< Service is redirecting requests to another node.
#define SP_SERVICE_STATE_STOPPED	0x4	///< Service is stopped.

struct SP_SERVICE_outgoing_replication;
struct SP_SERVICE_incoming_replication;

/**
 * @brief Service data structure.
 */
typedef struct {
	char					id[SP_SERVICE_ID_LENGTH];		///< Service ID.
	u_int8_t				state;					///< Current state.
	char*					data;					///< Pointer to service data.
	size_t					data_size;				///< Size of service data in bytes.
	struct SP_SERVICE_outgoing_replication*	outgoing_replication;			///< Pointer to outgoing replication.
	struct SP_SERVICE_incoming_replication*	incoming_replication;			///< Pointer to incoming replication.
	SP_addr_t				redirect_addr;				///< Address of node to redirect requests to.
	SP_PLACEMENT_service_stats_t*		placement_stats;			///< Statistics for service placement.
} SP_SERVICE_service_t;

/**
 * @brief Entry in the service cache.
 */
typedef struct {
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	SP_addr_t				server;					///< Address of the server.
	u_int16_t				age;					///< Age of this entry.
	bool					reachable;				///< Flag whether this entry is reachable.
} SP_SERVICE_service_cache_entry_t;

/**
 * @brief Data request data structure.
 */
typedef struct {
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	u_int16_t				offset;					///< Offset in service data.
	size_t					size;					///< Request size in bytes.
} SP_SERVICE_data_request_t;

/**
 * @brief Data update data structure.
 */
typedef struct {
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	u_int16_t				offset;					///< Offset in service data.
	size_t					size;					///< Update size in bytes.
	const char*				data;					///< New service data.
} SP_SERVICE_data_update_t;

/**
 * @brief Stream request data structure.
 */
typedef struct {
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	u_int16_t				offset;					///< Offset in service data.
	size_t					size;					///< Request size in bytes.
	SP_time_t				interval;				///< Request interval in seconds.
	SP_time_t				duration;				///< Request duration in seconds.
} SP_SERVICE_stream_request_t;

/**
 * @brief Active stream being served.
 */
typedef struct {
	SP_addr_t				to;					///< Address of the destination node.
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	u_int16_t				offset;					///< Offset in service data.
	size_t					size;					///< Request size in bytes.
	SP_time_t				interval;				///< Request interval in seconds.
	SP_time_t				end_time;				///< Request duration in seconds.
} SP_SERVICE_stream_t;

/**
 * @brief Segment of service data.
 */
typedef struct {
	u_int16_t				offset;					///< Segment offset.
	size_t					size;					///< Segment size.
} SP_SERVICE_segment_t;

/**
 * @brief Service replication result.
 */
typedef struct {
	SP_errno_t				errno;					///< Error code.
	SP_callback_parameter_t			parameter;				///< Parameter for callback function.
} SP_SERVICE_replication_result_t;

/**
 * @brief Outgoing service replication.
 */
typedef struct SP_SERVICE_outgoing_replication {
	SP_SERVICE_service_t*			service;				///< Pointer to service.
	SP_addr_t				destination;				///< Address of destination node.
	SP_callback_function_t			callback;				///< Callback function.
	SP_SERVICE_replication_result_t*	result;					///< Pointer to replication result.
} SP_SERVICE_outgoing_replication_t;

/**
 * @brief Incoming service replication.
 */
typedef struct SP_SERVICE_incoming_replication {
	SP_SERVICE_service_t*			service;				///< Pointer to service.
	SP_addr_t				source;					///< Address of source node.
	SP_list_t*				missing_segments;			///< List of missing service data segments.
} SP_SERVICE_incoming_replication_t;

/**
 * @brief Service migration result.
 */
typedef struct {
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	SP_errno_t				errno;					///< Error code.
	SP_callback_parameter_t			parameter;				///< Parameter for callback function.
} SP_SERVICE_migration_result_t;

/**
 * @brief Service migration.
 */
typedef struct {
	SP_SERVICE_service_t*			service;				///< Pointer to service.
	SP_addr_t				destination;				///< Address of destination node.
	SP_callback_function_t			callback;				///< Callback function.
	SP_SERVICE_migration_result_t*		result;					///< Pointer to migration result.
} SP_SERVICE_migration_t;



/*
 * Packet types.
 */

/**
 * @brief SERVICE DATA REQUEST packet.
 */
typedef struct {
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	u_int16_t				offset;					///< Offset in service data.
	size_t					size;					///< Request size in bytes.
	SP_time_t				start_time;				///< Start time of request.
} SP_SERVICE_CLIENT_DATA_REQUEST_packet_t;

/**
 * @brief SERVICE DATA UPDATE packet header.
 */
typedef struct {
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	u_int16_t				offset;					///< Offset in service data.
} SP_SERVICE_CLIENT_DATA_UPDATE_packet_t;

/**
 * @brief SERVICE STREAM REQUEST packet.
 */
typedef SP_SERVICE_stream_request_t SP_SERVICE_CLIENT_STREAM_REQUEST_packet_t;

/**
 * @brief SERVICE DATA ITEM packet.
 */
typedef struct {
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	SP_time_t				start_time;				///< Start time of request.
	SP_addr_t				redirect_addr;				///< Address of new server, if service is redirecting.
} SP_SERVICE_SERVER_DATA_ITEM_packet_t;

/**
 * @brief SERVICE STREAM ITEM packet.
 */
typedef struct {
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	SP_addr_t				redirect_addr;				///< Address of new server, if service is redirecting.
} SP_SERVICE_SERVER_STREAM_ITEM_packet_t;

/**
 * @brief SERVICE SERVER NOT AVAILABLE packet.
 */
typedef struct {
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	SP_addr_t				redirect_addr;				///< Address of new server, if service is redirecting.
} SP_SERVICE_SERVER_NOT_AVAILABLE_packet_t;

/**
 * @brief SERVICE REPLICATION REQUEST packet.
 */
typedef struct {
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	size_t					service_data_size;			///< Size of service data in bytes.
} SP_SERVICE_REPLICATION_REQUEST_packet_t;

/**
 * @brief SERVICE REPLICATION DATA packet.
 */
typedef struct {
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	u_int16_t				segment_offset;				///< Offset of service data segment.
} SP_SERVICE_REPLICATION_DATA_packet_t;

/**
 * @brief SERVICE REPLICATION ACK packet.
 */
typedef struct {
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	SP_SERVICE_segment_t			next_segment;				///< Next segment to transmit.
} SP_SERVICE_REPLICATION_ACK_packet_t;

/**
 * @brief SERVICE REPLICATION NACK packet.
 */
typedef struct {
	char					service_id[SP_SERVICE_ID_LENGTH];	///< Service ID.
	SP_errno_t				error_code;				///< Error code.
} SP_SERVICE_REPLICATION_NACK_packet_t;

#endif	// SERVICE_TYPES_H
