/**
 * @file types.h
 *
 * @brief Types and macro definitions.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: types.h 2548 2008-02-13 19:20:58Z wittenbu $
 */

#ifndef TYPES_H
#define TYPES_H



#ifdef SP_ARCH_NONE
	// Used when generating dependencies.
#elif defined(SP_ARCH_LINUX)
	#include "arch/linux/types.h"
#elif defined(SP_ARCH_NS2)
	// Nothing to do.
#else
	#error No architecture defined (available architectures are: SP_ARCH_LINUX and SP_ARCH_NS2).
#endif	// SP_ARCH_*



#include <sys/types.h>
#include <stdlib.h>
#include <string.h>



/**
 * @brief Wrapper macro around malloc().
 *
 * Allocates memory and initializes it to 0x0. Aborts if no memory was
 * available.
 *
 * @param p Pointer at which memory is to be allocated.
 * @param type Type of data structure.
 */
#define SP_MALLOC(p, type) \
	p = (type*) malloc(sizeof(type)); \
	if(p == NULL) { \
		printf("Unable to allocate memory.\n"); \
		abort(); \
	} \
	memset(p, 0, sizeof(type))

/**
 * @brief Wrapper macro around malloc() for allocating arrays.
 *
 * @param p Pointer at which memory is to be allocated.
 * @param type Type of data structure.
 * @param count Number of items of the data structure that are to fit into the
 * allocated memory.
 */
#define SP_MALLOC_ARRAY(p, type, count) \
	p = (type*) malloc(sizeof(type) * count); \
	if(p == NULL) { \
		printf("Unable to allocate memory.\n"); \
		abort(); \
	} \
	memset(p, 0, sizeof(type) * count)

/**
 * @brief Wrapper macro around free().
 *
 * Frees memory and sets the pointer to NULL.
 *
 * @param p Pointer to memory.
 */
#define SP_FREE(p) \
	free(p); \
	p = NULL

/**
 * @brief Join two memory regions.
 *
 * Join two memory regions into a newly allocated memory region of the combined
 * size.
 *
 * @param p Pointer at which memory is to be allocated.
 * @param type Type of data structure.
 * @param p1 Pointer to first memory region.
 * @param size1 Size of first memory region.
 * @param p2 Pointer to second memory region.
 * @param size2 Size of second memory region.
 */
#define SP_MALLOC_JOIN(p, type, p1, size1, p2, size2) \
	if(sizeof(type) > (size1 + size2)) { \
		printf("Size of type of joined memory region is greater than allocated memory.\n"); \
		abort(); \
	} \
	char* tmp; \
	SP_MALLOC_ARRAY(tmp, char, size1 + size2); \
	p = (type*) tmp; \
	memcpy(p, p1, size1); \
	memcpy(((char*) p) + size1, p2, size2);

/**
 * @brief Split the content of a memory region into two already allocated regions.
 *
 * @param p1 Pointer to first memory region.
 * @param size1 Size of first memory region.
 * @param p2 Pointer to second memory region.
 * @param size2 Size of second memory region.
 * @param p Pointer to memory that is to be split.
 */
#define SP_SPLIT(p1, size1, p2, size2, p) \
	memcpy(p1, p, size1); \
	memcpy(p2, ((char*) p) + size1, size2)



/**
 * @brief Item of a single-linked list.
 */
typedef struct SP_list_item {
	void*			data;		///< Pointer to data of this list item.
	struct SP_list_item*	next;		///< Pointer to next list item.
} SP_list_item_t;

/**
 * @brief Single-linked list with some extra fields for efficiency.
 */
typedef struct {
	SP_list_item_t*		first_item;	///< Pointer to first item of the list.
	SP_list_item_t*		last_item;	///< Pointer to last item of the list.
	u_int16_t 		size;		///< Number of items in the list.
} SP_list_t;



/**
 * @brief Callback parameter type.
 */
typedef void* SP_callback_parameter_t;

/**
 * @brief Callback function type.
 */
typedef void(*SP_callback_function_t)(SP_callback_parameter_t);

/**
 * @brief Callback consisting of function, parameter and time.
 */
typedef struct {
	SP_callback_function_t	function;	///< Callback function.
	SP_callback_parameter_t parameter;	///< Callback parameter.
	SP_time_t		time;		///< Time at which the callback function will be called.
} SP_callback_t;

/**
 * @brief Callback error code function type.
 */
typedef u_int8_t SP_errno_t;



#define SP_APP_NONE		0x0	///< Packets for no component (must not happen).
#define SP_APP_ICMP		0x1	///< Packets for the ICMP component.
#define SP_APP_RUDP		0x2	///< Packets for the RUDP component.
#define SP_APP_SLP		0x3	///< Packets for the SLP component.
#define SP_APP_SERVICE		0x4	///< Packets for the SERVICE component.
#define SP_APP_PLACEMENT	0x5	///< Packets for the PLACEMENT component.

#define SP_TYPE_NONE		0x0	///< Packets of no type (must not happen)

/**
 * @brief Time-to-live (TTL) type.
 */
typedef u_int8_t SP_ttl_t;

/**
 * @brief Sequence number type.
 */
typedef u_int16_t SP_seqno_t;

/**
 * @brief Packet header.
 */
typedef struct {
	SP_addr_t	from;			///< Source address.
	SP_addr_t	to;			///< Destination address.
	u_int8_t	hops;			///< Number of times this packet was forwarded.
	SP_addr_t	last_hop;		///< Address of last node that forwarded this packet.
	SP_ttl_t	ttl;			///< Time-to-live (TTL).
	SP_seqno_t	seqno;			///< Sequence number.
	u_int8_t	app;			///< Application identifier.
	u_int8_t	type;			///< Application-specific packet type identifier.
	size_t		payload_length;		///< Size of the payload in bytes.
} SP_packet_header_t;

/**
 * @brief Payload type.
 */
typedef unsigned char SP_packet_payload_t;

/**
 * @brief Packet consisting of header and payload.
 */
typedef struct {
	SP_packet_header_t*	header;		///< Pointer to packet header.
	SP_packet_payload_t*	payload;	///< Pointer to packet payload.
} SP_packet_t;



/**
 * @brief Packet fingerprint.
 *
 * This data structure is used for checking for duplicates in recently arrived
 * packets.
 */
typedef struct {
	SP_time_t	time;			///< Time of arrival.
	SP_addr_t	from;			///< Source address.
	SP_seqno_t	seqno;			///< Sequence number.
} SP_packet_fingerprint_t;



/**
 * @brief Basic information about a node in the network.
 */
typedef struct {
	SP_addr_t	address;		///< Node address.
	u_int8_t	hops;			///< Distance in hops.
	SP_addr_t	last_hop;		///< Address of last hop.
	u_int32_t	packets;		///< Number of packtes received from this node.
} SP_node_t;

#endif	// TYPES_H
