/**
 * @file debug.c
 *
 * @brief Debugging functions.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: debug.c 2563 2008-02-19 14:23:37Z wittenbu $
 */

#include <execinfo.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>

#include "../types.h"
#include "../vars.h"
#include "../interface.h"
#include "../interface_internal.h"
#include "debug.h"
#include "net.h"



#ifdef SP_ENABLE_DEBUG

/**
 * @brief Use color in output.
 */
#define DEBUG_USE_COLOR

/**
 * @brief Size of backtrace on FATAL errors.
 */
#define DEBUG_BACKTRACE_SIZE	64

/**
 * @brief Convert numeric debug level to a string.
 *
 * @param debug_level Numeric debug level.
 * @return A string representing the debug level.
 */
static const char* SP_DEBUG_LEVEL_TO_STRING(int debug_level) {
	switch(debug_level) {
		case SP_DEBUG_LEVEL_NONE:	SP_printf("FATAL: Formatted output for debug level NONE requested.\n"); abort();
		case SP_DEBUG_LEVEL_FATAL:	return "FATAL";
		case SP_DEBUG_LEVEL_ERROR:	return "ERROR";
		case SP_DEBUG_LEVEL_WARN:	return "WARN";
		case SP_DEBUG_LEVEL_NOTE:	return "NOTE";
		case SP_DEBUG_LEVEL_INFO:	return "INFO";
		case SP_DEBUG_LEVEL_DEBUG:	return "DEBUG";
		case SP_DEBUG_LEVEL_VERBOSE:	return "VERB";
		default: SP_printf("FATAL: Unknown debug level 0x%X.\n", debug_level); abort();
	}
}

#ifdef DEBUG_USE_COLOR
/**
 * @brief Return ASCII control sequence to highlight debugging output.
 *
 * @param debug_level Numeric debug level.
 * @return A string with a ASCII control sequence for appropriate highlighting.
 */
static const char* SP_DEBUG_LEVEL_TO_COLOR(int debug_level) {
	switch(debug_level) {
		case SP_DEBUG_LEVEL_NONE:	return "\033[0m";		// reset to default
		case SP_DEBUG_LEVEL_FATAL:	return "\033[41m\033[1;31m";	// bright red on red
		case SP_DEBUG_LEVEL_ERROR:	return "\033[40m\033[1;31m";	// bright red on black
		case SP_DEBUG_LEVEL_WARN:	return "\033[40m\033[1;33m";	// yellow on black
		case SP_DEBUG_LEVEL_NOTE:	return "\033[40m\033[1;32m";	// green on black
		case SP_DEBUG_LEVEL_INFO:	return "\033[40m\033[1;34m";	// blue on black
		case SP_DEBUG_LEVEL_DEBUG:	return "\033[40m\033[1;37m";	// light gray on black
		case SP_DEBUG_LEVEL_VERBOSE:	return "\033[40m\033[1;30m";	// dark gray on black
		default: SP_printf("FATAL: Unknown debug level 0x%X.\n", debug_level); abort();
	}
}
#endif	// DEBUG_USE_COLOR

#endif	// SP_ENABLE_DEBUG

/**
 * @brief Print formatted debuggging output.
 *
 * @param debug_level Numeric debug level of the output.
 * @param format A format string.
 * @param ... Parameters for the format string.
 */
void debug_printf(int debug_level, const char *format, ...) {
#ifdef SP_ENABLE_DEBUG
	if(debug_level == SP_DEBUG_LEVEL_ERROR) {
		SP_DEBUG_ERRORs++;
	} else if(debug_level == SP_DEBUG_LEVEL_WARN) {
		SP_DEBUG_WARNs++;
	}
	if(debug_level == SP_DEBUG_LEVEL_NONE) {
		va_list argp;
		va_start(argp, format);
		SP_vprintf(format, argp);
		va_end(argp);
	} else if(debug_level <= SP_DEBUG_LEVEL) {
#ifdef DEBUG_USE_COLOR
		SP_printf("%s", SP_DEBUG_LEVEL_TO_COLOR(debug_level));
#endif	// DEBUG_USE_COLOR
#ifdef SP_ARCH_LINUX
		time_t time = (time_t) SP_time();
		char time_string[9];
		strftime(time_string, 9, "%T", localtime(&time));
		SP_printf("[%s @ %s: %s]", time_string, addr_to_str(SP_LOCAL_ADDR), SP_DEBUG_LEVEL_TO_STRING(debug_level));
#else
		SP_printf("[%.2f @ %s: %s]", SP_time(), addr_to_str(SP_LOCAL_ADDR), SP_DEBUG_LEVEL_TO_STRING(debug_level));
#endif	// SP_ARCH_*
#ifdef DEBUG_USE_COLOR
		SP_printf("%s", SP_DEBUG_LEVEL_TO_COLOR(SP_DEBUG_LEVEL_NONE));
#endif	// DEBUG_USE_COLOR
		SP_printf("\t");
		va_list argp;
		va_start(argp, format);
		SP_vprintf(format, argp);
		va_end(argp);
		if(debug_level == SP_DEBUG_LEVEL_FATAL) {
			SP_printf("\n");
			SP_printf("\t\t\t\t\\|/ ____ \\|/\n");
			SP_printf("\t\t\t\t\"@'/ .. \\`@\"\n");
			SP_printf("\t\t\t\t/_| \\__/ |_\\\n");
   			SP_printf("\t\t\t\t   \\__U_/\n");
			SP_printf("\n");

			// Print backtrace.
			void *array[DEBUG_BACKTRACE_SIZE];
			size_t size = backtrace(array, DEBUG_BACKTRACE_SIZE);
			char **strings = backtrace_symbols(array, size);
			size_t i;
			SP_printf("Backtrace of %zd stack frame(s):\n", size);
			for(i = 0; i < size; i++)
				SP_printf("%s\n", strings[i]);
			free (strings);
			SP_printf("\n");
		}
	}
#endif	// SP_ENABLE_DEBUG
}
