/**
 * @file net.c
 *
 * @brief Network utility functions.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: net.c 2548 2008-02-13 19:20:58Z wittenbu $
 */

#include <string.h>
#include <stdlib.h>

#if SP_ARCH == LINUX
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#endif	// SP_ARCH == LINUX

#include "../types.h"
#include "../vars.h"
#include "../interface.h"
#include "../interface_internal.h"
#include "debug.h"
#include "list.h"
#include "callback.h"
#include "net.h"



/**
 * @brief Convert a numeric node address to a string.
 *
 * This function may be called several times in function parameters, so we have
 * to be careful to use different buffers every time.
 *
 * @param addr Numeric node address.
 * @return String representation of the node address.
 */
char* addr_to_str(SP_addr_t addr) {
#ifdef SP_ARCH_LINUX
	struct in_addr tmp_addr;
	tmp_addr.s_addr = addr;
	return inet_ntoa(tmp_addr);
#else
	int index = ++SP_addr_to_str_buffer_index % SP_ADDR_TO_STR_BUFFER_SIZE;
	snprintf(SP_addr_to_str_buffer[index], SP_ADDR_TO_STR_BUFFER_STR_LEN, "#%d", addr);
	return SP_addr_to_str_buffer[index];
#endif	// SP_ARCH_*
}



/**
 * @brief Generate random node address excluding the local address.
 *
 * @return Random node address excluding the local address.
 */
SP_addr_t random_node_addr() {
	SP_addr_t addr = (SP_addr_t) SP_random(SP_NUMBER_OF_NODES);
	while(addr == SP_LOCAL_ADDR) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "Random node address matches local address. Generating new one...\n");
		addr = (SP_addr_t) SP_random(SP_NUMBER_OF_NODES);
	}
	return addr;
}



/**
 * @brief Initialize of packet header.
 *
 * @param header Pointer to packet header to initialize.
 */
void init_header(SP_packet_header_t* header) {
	SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Initializing packet header at %p...\n", header);
	header->from = SP_LOCAL_ADDR;
	header->to = SP_BROADCAST_ADDR;
	header->hops = 0;
	header->ttl = SP_DEFAULT_TTL;
	header->seqno = SP_sequence_number++;
	header->app = SP_APP_NONE;
	header->type = SP_TYPE_NONE;
	header->payload_length = 0;
}



/**
 * @brief Maximal randomized delay before forwarding a packet in seconds.
 *
 * Uncommenting this define causes packet to be forwarded immediatly.
 */
#define RUDP_BROADCAST_FORWARD_DELAY	0
//#define RUDP_BROADCAST_FORWARD_DELAY	SP_EXPECTED_ONE_HOP_RTT

/**
 * @brief Packet forwarding callback to support delayed forwarding.
 *
 * @param parameter Pointer to packet to be forwarded.
 */
static void forward_callback(SP_callback_parameter_t parameter) {
	SP_packet_t* packet = (SP_packet_t*) parameter;
	SP_send(packet->header, packet->payload);
	if(packet->header->payload_length > 0)
		SP_FREE(packet->payload);
	SP_FREE(packet->header);
	SP_FREE(packet);
}

/**
 * @brief Forward a packet.
 *
 * Re-sends packets with non-zero TTL. Decreases TTL before sending.
 *
 * @param header Pointer to header of packet to be forwarded.
 * @param payload Pointer to payload of packet to be forwarded.
 */
void forward(const SP_packet_header_t* header, const SP_packet_payload_t* payload) {
	SP_packet_header_t new_header;
	if(header->to == SP_LOCAL_ADDR) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "Dropping packet instead of forwarding it to the local node...\n");
	} else if(header->ttl > 1) {
		SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Forwarding packet...\n");
		memcpy(&new_header, header, sizeof(SP_packet_header_t));
		new_header.ttl--;
		if(RUDP_BROADCAST_FORWARD_DELAY == 0) {
			SP_send(&new_header, payload);
		} else {
			if(new_header.to != SP_BROADCAST_ADDR) {
				SP_send(&new_header, payload);
			} else {		// Add delay to broadcast packets.
				SP_packet_t* packet;
				SP_MALLOC(packet, SP_packet_t);
				SP_MALLOC(packet->header, SP_packet_header_t);
				memcpy(packet->header, &new_header, sizeof(SP_packet_header_t));
				if(packet->header->payload_length > 0) {
					char* p;
					SP_MALLOC_ARRAY(p, char, packet->header->payload_length);
					memcpy(p, payload, packet->header->payload_length);
					packet->payload = (SP_packet_payload_t*) p;
				} else {
					packet->payload = NULL;
				}
				callback_add(forward_callback, packet, SP_random(RUDP_BROADCAST_FORWARD_DELAY));
			}
		}
	} else {
		SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Dropping packet with expired TTL...\n");
	}
}



/**
 * @brief Check for reception of duplicate packets.
 *
 * Check whether a packet has already been received recently and add it to the
 * corresponding data structure for future checks.
 *
 * @param header Pointer to header of a recently received packet.
 * @return <code>true</code> if the packet is a duplicate based on its
 * fingerprint, <code>false</code> otherwise.
 */
bool check_for_duplicate(const SP_packet_header_t* header) {
	int i, LRU_index = -1;
	SP_time_t current_time = SP_time();
	SP_time_t time_of_oldest_item = current_time;
	// Do a reverse loop to use the beginning of the cache first and ease debugging.
	for(i = SP_RECEIVED_PACKET_CACHE_SIZE - 1; i >= 0; i--) {
		// Check whether we have recently seen a similar packet.
		if(
			(SP_received_packet_cache[i].time != 0)
			&& (SP_received_packet_cache[i].from == header->from)
			&& (SP_received_packet_cache[i].seqno == header->seqno)
		) {
			SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Duplicate packet fingerprint (time=%.2f, from=%s, seqno=%d) found at index %d.\n", current_time, addr_to_str(header->from), header->seqno, i);
			SP_received_packet_cache[i].time = current_time;
			return true;
		}
		// Just in case, find the least recently used entry in data
		// structure for storing information about this packet.
		if(
			(SP_received_packet_cache[i].time == 0)
			|| (SP_received_packet_cache[i].time < time_of_oldest_item)
		) {
			time_of_oldest_item = SP_received_packet_cache[i].time;
			LRU_index = i;
		}
	}
	if(LRU_index < 0) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Failed to find slot for packet fingerprint (time=%f, from=%s, seqno=%d).\n", current_time, addr_to_str(header->from), header->seqno);
		abort();
	}
	SP_received_packet_cache[LRU_index].time = current_time;
	SP_received_packet_cache[LRU_index].from = header->from;
	SP_received_packet_cache[LRU_index].seqno = header->seqno;
	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "Added packet fingerprint (time=%f, from=%s, seqno=%d) at index %d.\n", SP_received_packet_cache[LRU_index].time, addr_to_str(SP_received_packet_cache[LRU_index].from), SP_received_packet_cache[LRU_index].seqno, LRU_index);
	return false;
}



/**
 * @brief Check whether two nodes have matching addresses.
 *
 * @param item Node to check.
 * @param cmp_item Reference node.
 * @return <code>true</code> if the two nodes have the same address,
 * <code>false</code> otherwise.
 */
static bool matches_node_addr(const void* item, const void* cmp_item) {
	SP_node_t* node = (SP_node_t*) item;
	SP_node_t* cmp_node = (SP_node_t*) cmp_item;
	return (node->address == cmp_node->address);
}

/**
 * @brief Adds or updates a node in the network map.
 *
 * @param address Address of the node.
 * @param hops Distance to the node.
 * @param last_hop Address of the node that is the last hop on the route.
 */
void add_node_to_network_map(SP_addr_t address, u_int8_t hops, SP_addr_t last_hop) {
	if(address == SP_LOCAL_ADDR)
		return;

	SP_node_t cmp_node;
	cmp_node.address = address;
	SP_node_t* node = (SP_node_t*) list_find(SP_network_map, &cmp_node, matches_node_addr);
	if(node != NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Updating node %s at %d hop(s) with last hop %s in network map...\n", addr_to_str(address), hops, addr_to_str(last_hop));
		node->hops = hops;
		node->last_hop = last_hop;
		node->packets++;
	} else {
		SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Adding node %s at %d hop(s) with last hop %s to network map...\n", addr_to_str(address), hops, addr_to_str(last_hop));
		SP_MALLOC(node, SP_node_t);
		node->address = address;
		node->hops = hops;
		node->last_hop = last_hop;
		node->packets = 1;
		list_add(SP_network_map, node);
	}
}

/**
 * @brief Update the aggregate network statistics counters.
 */
void calculate_network_statistics() {
	u_int32_t direct_neighbors = 0;
	u_int32_t longest_route = 0;
	double total_route_length = 0;
	unsigned int i;
	SP_node_t* node;
	for(i = 0; i < list_size(SP_network_map); i++) {
		node = (SP_node_t*) list_get(SP_network_map, i);
		if(node->hops == 0) {
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Node %s has hop count 0, last node was %s.\n", addr_to_str(node->address), addr_to_str(node->last_hop));
			abort();
		} else if(node->hops == 1) {
			direct_neighbors++;
		}
		if(node->hops > longest_route)
			longest_route = node->hops;
		total_route_length += node->hops;
	}
	SP_direct_neighbors = direct_neighbors;
	SP_longest_route = longest_route;
	if(list_size(SP_network_map) > 0)
		SP_average_route = total_route_length / list_size(SP_network_map);
}
