/**
 * @file vars.h
 *
 * @brief Global variables and constants.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: vars.h 2569 2008-02-20 17:24:59Z wittenbu $
 */

#ifndef VARS_H
#define VARS_H

#include <stdio.h>
#include <stdlib.h>

#include "types.h"



/**
 * @brief Global constant to enable debugging.
 */
#define SP_ENABLE_DEBUG

#ifdef SP_ENABLE_DEBUG
	#define SP_DEBUG_LEVEL_NONE		0x0	///< No debug level given.
	#define SP_DEBUG_LEVEL_FATAL		0x1	///< Debug level for fatal errors.
	#define SP_DEBUG_LEVEL_ERROR		0x2	///< Debug level for regular errors.
	#define SP_DEBUG_LEVEL_WARN		0x3	///< Debug level for warnings.
	#define SP_DEBUG_LEVEL_NOTE		0x4	///< Debug level for important notes.
	#define SP_DEBUG_LEVEL_INFO		0x5	///< Debug level for regular information.
	#define SP_DEBUG_LEVEL_DEBUG		0x6	///< Debug level for debug output.
	#define SP_DEBUG_LEVEL_VERBOSE		0x7	///< Debug level for verbose debug output.

	/**
	 * @brief Current debug level.
	 */
	#define SP_DEBUG_LEVEL			SP_DEBUG_LEVEL_NOTE

	/**
	 * @brief Debug output function macro (enabled)
	 */
	#define SP_DEBUG			debug_printf
#else
	/**
	 * @brief Debug output function macro (disabled)
	 */
	#define SP_DEBUG			if(0) debug_printf
#endif	// SP_ENABLE_DEBUG



#define SP_ERROR_NOT_SET		0x0	///< Error code not set (must not happen).
#define SP_ERROR_OK			0x1	///< No error, everything OK.
#define SP_ERROR_PARAMETER		0x2	///< Error due to invalid parameter.
#define SP_ERROR_NOT_FOUND		0x3	///< Error due to requested resource not being found.
#define SP_ERROR_TIMEOUT		0x4	///< Error due to timeout.
#define SP_ERROR_OUT_OF_MEMORY		0x5	///< Error due to lack of memory.
#define SP_ERROR_BUSY			0x6	///< Error due to a component being busy.
#define SP_ERROR_IN_PROGRESS		0x7	///< Error due to a similar request already being in progress.
#define SP_ERROR_COMM_ERROR		0x8	///< Error due to a communication error.
#define SP_ERROR_REMOTE_BUSY		0x9	///< Error due to the remote component / node being busy.



/**
 * @brief List of pending callbacks.
 */
SP_list_t* SP_callbacks;

/**
 * @brief Time of next pending callback.
 */
SP_time_t SP_callback_time;

/**
 * @brief Number of registered callbacks.
 */
u_int32_t SP_registered_callbacks;

/**
 * @brief Number of executed callbacks.
 */
u_int32_t SP_executed_callbacks;



/**
 * @brief Convenience macro for address of the local node.
 */
#define SP_LOCAL_ADDR			SP_local_addr

/**
 * @brief Address of the local node.
 */
SP_addr_t SP_local_addr;



#ifdef SP_ARCH_LINUX
	/**
	 * @brief Default UDP port number.
	 */
	#define SP_DEFAULT_UDP_PORT	8380

	/**
	 * @brief UDP socket handle.
	 */
	int SP_udp_socket;
#endif	// SP_ARCH_LINUX



/**
 * @brief Maximum transmission unit (MTU) in bytes.
 */
#define SP_MTU				1500

/**
 * @brief Expected network diameter.
 */
#define SP_EXPECTED_NETWORK_DIAMETER	32

/**
 * @brief Default time-to-live (TTL) for transmitted packets.
 */
#define SP_DEFAULT_TTL			SP_EXPECTED_NETWORK_DIAMETER

/**
 * @brief Expected single-hop round trip time.
 *
 * Expected round trip time (RTT) for single-hop packet transmission in
 * seconds.
 */
#define SP_EXPECTED_SINGLE_HOP_RTT	2.0

/**
 * @brief Expected multi-hop round trip time.
 *
 * Expected round trip time (RTT) for multi-hop packet transmission in
 * seconds. This value must be greater than 1.0 because it is used as base when
 * calculating the exponential backoff time.
 */
#define SP_EXPECTED_MULTI_HOP_RTT	(SP_EXPECTED_SINGLE_HOP_RTT * SP_EXPECTED_NETWORK_DIAMETER)

/**
 * @brief Expected time to establish a new route in seconds.
 */
#define SP_ROUTE_ESTABLISHMENT_TIME	(SP_EXPECTED_MULTI_HOP_RTT * 2.0)

/**
 * @brief Timeout for higherlevel protocols.
 *
 * It should allow for enough time for RUDP to transmit and acknowledge a
 * packet under normal network conditions.
 */
#define SP_TIMEOUT			(SP_ROUTE_ESTABLISHMENT_TIME + SP_EXPECTED_MULTI_HOP_RTT * 3.0)

/**
 * @brief Current sequence number for transmitted packets.
 */
u_int32_t SP_sequence_number;

/**
 * @brief Size of cache for recently received packets.
 */
#define SP_RECEIVED_PACKET_CACHE_SIZE	64

/**
 * @brief Cache for recently received packets.
 */
SP_packet_fingerprint_t SP_received_packet_cache[SP_RECEIVED_PACKET_CACHE_SIZE];



/**
 * @brief Network map.
 *
 * This list is to collect information about all nodes that this node has been
 * in contact with.
 */
SP_list_t* SP_network_map;

/**
 * @brief Number of direct (one-hop) neighbors of this node.
 *
 * This variable is only updated on demand.
 */
u_int32_t SP_direct_neighbors;

/**
 * @brief Length of the longest route leading to this node in hops.
 *
 * This variable is only updated on demand.
 */
u_int32_t SP_longest_route;

/**
 * @brief Average length of all routes leading to this node.
 *
 * This variable is only updated on demand.
 */
double SP_average_route;



/**
 * @brief Maximal supported number of different services.
 *
 * Each service may have multiple instances.
 */
#define SP_MAX_SERVICES	4



/**
 * @brief Size of buffer for converting numeric addresses to a string
 * representation.
 */
#define SP_ADDR_TO_STR_BUFFER_SIZE	4

/**
 * @brief Length of strings in buffer for converting numeric addresses to a
 * string representation.
 */
#define SP_ADDR_TO_STR_BUFFER_STR_LEN	5

/**
 * @brief Buffer for converting numeric addresses to a string representation.
 *
 * Multiple addresses may need to be valid at the same point in time. Hence, we
 * need a buffer with multiple slots to avoid overwriting strings that still
 * need to be processed.
 */
char SP_addr_to_str_buffer[SP_ADDR_TO_STR_BUFFER_SIZE][SP_ADDR_TO_STR_BUFFER_STR_LEN];

/**
 * @brief Currently used index of buffer for converting numeric addresses to a
 * string representation.
 */
u_int8_t SP_addr_to_str_buffer_index;



/**
 * @brief Number of ERRORs encountered during execution.
 */
u_int32_t SP_DEBUG_ERRORs;

/**
 * @brief Number of WARNings encountered during execution.
 */
u_int32_t SP_DEBUG_WARNs;

#endif	// VARS_H
